@@ -1,37 +1,47 @@
 import datetime
-import sys
 import pickle
 from os import makedirs
-import warnings
-if not sys.warnoptions:
-    warnings.simplefilter("ignore")
 
 from termcolor import colored
 from itertools import chain, combinations
 from multiprocessing import Pool, cpu_count
 
+import lightgbm as lgb
+from sklearn.model_selection import train_test_split
+from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
+from sklearn.metrics import f1_score, classification_report, confusion_matrix
+from seaborn import heatmap
+
+
+
+
+
+import sys
+import warnings
+if not sys.warnoptions:
+    warnings.simplefilter("ignore")
 
 import pandas as pd
 import numpy as np
 import matplotlib.pyplot as plt
-from seaborn import heatmap
-
-import lightgbm as lgb
-
-from sklearn.model_selection import train_test_split
-from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
-from sklearn.metrics import f1_score, classification_report, confusion_matrix
-
 
 from typing import Callable
 
 
-
 from . import nans
 from . import encoder
 from . import splitter
 from . import cross_validation
 from . import recover
+from . import fit_pred_bagged
+from . import feature_selection_bagged
+from . import save_load_algorithms
+
+
+
+
+
+
 
 
 
@@ -94,7 +104,6 @@ class Snaplib__222:
         Use case:
         df = Snaplib().encode_dataframe(df)
         '''
-
         if inplace is False:
             df_copied = df.copy()
 
@@ -114,7 +123,6 @@ class Snaplib__222:
         Use case:
         df = Snaplib().decode_dataframe(df)
         '''
-
         if inplace is False:
             df_copied = df.copy()
             return self.ENCODER.decode_dataframe(df_copied)
@@ -184,7 +192,8 @@ class Snaplib__222:
                                                     target_feature, 
                                                     test_size, 
                                                     random_state, 
-                                                    research)
+                                                    research
+                                                    )
 
 
 
@@ -198,7 +207,6 @@ class Snaplib__222:
                   verbose : int or bool = 0, 
                   early_stopping_rounds : int = 0, 
                   ) -> float:
-
         ''' 
         Cross Validation method for list of algorithms.
         
@@ -228,15 +236,16 @@ class Snaplib__222:
         cv is num K_FOLD integer 
         verbose = 0 mute, 1 verbose.
         early_stopping_rounds default 0 or positive int.
-
         '''
-        return cross_validation.cross_val(algorithms, 
-                                          k_fold_dict, 
-                                          metric, 
-                                          task, 
-                                          cv, 
-                                          verbose=verbose, 
-                                          early_stopping_rounds=early_stopping_rounds)
+        return cross_validation.cross_val(
+                    algorithms, 
+                    k_fold_dict, 
+                    metric, 
+                    task, 
+                    cv, 
+                    verbose=verbose, 
+                    early_stopping_rounds=early_stopping_rounds
+                    )
 
 
 
@@ -252,7 +261,7 @@ class Snaplib__222:
         Imputing of missing values (np.nan) in tabular data, not TimeSeries.
 
         Use case:
-        df = Snaplib().recover_data(df, verbose=True, stacking=True)
+        df = Snaplib().recover_data(df, verbose=True, discrete_columns="auto")
         device must be "cpu" or "gpu". Sometime small datasets work faster with cpu.
         if set verbose = if True algorithm runs cross validation tests and print results of tests for decision making.
         discrete_columns = ['col_name_1', 'col_name_2', 'col_name_3', 'etc']
@@ -269,6 +278,224 @@ class Snaplib__222:
 
 
 
+    @staticmethod
+    def fit_stacked(algorithms_list : list, 
+                    X_train : pd.DataFrame, 
+                    y_train : pd.Series or np.ndarray, 
+                    X_val : pd.DataFrame = None, 
+                    y_val : pd.Series or np.ndarray = None, 
+                    verbose : int or bool = 0, 
+                    early_stopping_rounds : int = 0, 
+                    ) -> list:
+        ''' 
+        Fit method for list of algorithms.
+        
+        Use case:
+        algorithms_list = Snaplib().fit_stacked(
+                                                algorithms_list, 
+                                                X_train, 
+                                                y_train, 
+                                                X_val=None, 
+                                                y_val=None,
+                                                verbose=0,
+                                                ):
+        
+        algorithms_list = list of algorithms [LGBMClassifier(), XGBClassifier(), CatBoostClassifier()].
+        X_train and y_train are data for training list of algorithms.
+
+        verbose = 0 mute, 1 verbose.
+        early_stopping_rounds default 0 or positive int.
+        '''
+        return  fit_pred_bagged.fit_stacked(
+                    algorithms_list, 
+                    X_train, 
+                    y_train, 
+                    X_val=X_val, 
+                    y_val=y_val, 
+                    verbose=verbose, 
+                    early_stopping_rounds=early_stopping_rounds, 
+                    )
+
+
+
+
+    @staticmethod
+    def predict_stacked(algorithms_list : list, 
+                        X_pred : pd.DataFrame, 
+                        task : str = 'clsf'
+                        ) -> list:
+        ''' 
+        Prediction method for list of algorithms.
+        
+        Use case:
+        y_hat = predict_stacked(self, 
+                                algorithms_list, 
+                                X, 
+                                task='clsf'
+                                ):
+        
+        algorithms_list = list of algorithms [LGBMClassifier(), XGBClassifier(), CatBoostClassifier()].
+        X_pred is dataframe for prediction.
+        task='clsf' or 'regr', classification or regression
+        '''
+        return  fit_pred_bagged.predict_stacked(algorithms_list, 
+                                                X_pred, 
+                                                task=task,
+                                                )
+
+
+
+
+    @staticmethod
+    def features_selection_regr(algorithms : list, 
+                                df : pd.DataFrame, 
+                                target : str, 
+                                metric : Callable, 
+                                cv : int, 
+                                verbose : int or bool = 0, 
+                                early_stopping_rounds : int = 0, 
+                                ) -> list:
+        '''
+        Select bests features for modeling and return list with bad features required to be droped.
+
+        Use case:
+        features_to_drop = Snaplib().features_selection_regr(algorithms, df, target, metric, cv, verbose=0):
+        df.drop(features_to_drop, inplace=True, axis=1)
+
+
+        algorithms is a list of algorithms like algs = [
+                                                        [LGBMRegressor, dict(params)],
+                                                        [XGBRegressor, dict(params)], 
+                                                        [CatBoostRegressor, dict(params)],
+                                                        ]
+
+        df = pandas.core.frame.DataFrame.
+        target = name of target of str type.
+
+        metric is a metric like f1_score or mean_absolute_error.
+        cv is num K_FOLD integer 
+        verbose = 0 mute, 1 verbose.
+        early_stopping_rounds default 0 or positive int.
+        '''
+        return feature_selection_bagged.features_selection_regr(
+                    algorithms, 
+                    df, 
+                    target, 
+                    metric, 
+                    cv, 
+                    verbose=verbose, 
+                    early_stopping_rounds=early_stopping_rounds, 
+                    )
+
+
+
+
+    @staticmethod
+    def features_selection_clsf(algorithms : list, 
+                                df : pd.DataFrame, 
+                                target : str, 
+                                metric : Callable, 
+                                cv : int, 
+                                verbose : int or bool = 0, 
+                                early_stopping_rounds : int = 0, 
+                                ) -> list:
+        ''' 
+        Select bests features for modeling and return list with bad features required to be droped.
+        
+        Use case:
+        features_to_drop = Snaplib().features_selection_clsf(algorithms, df, target, metric, cv, verbose=0):
+        df.drop(features_to_drop, inplace=True, axis=1)
+
+
+        algorithms is a list of algorithms like algs = [
+                                                        [LGBMClassifier, dict(params)],
+                                                        [XGBClassifier, dict(params)], 
+                                                        [CatBoostClassifier, dict(params)],
+                                                        ]
+
+        df = pandas.core.frame.DataFrame.
+        target = name of target of str type.
+                
+        metric is a metric like f1_score or mean_absolute_error.
+        cv is num K_FOLD integer 
+        verbose = 0 mute, 1 verbose.
+        early_stopping_rounds default 0 or positive int.
+        '''
+        return feature_selection_bagged.features_selection_clsf(
+                    algorithms, 
+                    df, 
+                    target, 
+                    metric, 
+                    cv, 
+                    verbose=verbose, 
+                    early_stopping_rounds=early_stopping_rounds, 
+                    )
+
+
+
+
+
+    @staticmethod
+    def save_stack( algorithms_list : list, 
+                    directory : str = ''
+                    ) -> None:
+        ''' 
+        Save method for all in list of algorithms in directory.
+        Return list of file names
+        
+        Use case:
+        file_names = save_stack(algorithms_list, directory='')
+        
+        algorithms_list = list of algorithms [LGBMClassifier(), XGBClassifier(), CatBoostClassifier()].
+        directory:
+        '' - save files to current working directory
+        or
+        save files to /some/directory/not/exist/
+        '''
+        return save_load_algorithms.save_stack(algorithms_list, 
+                                               directory=directory)
+
+
+
+
+    @staticmethod
+    def load_stack( names_list : list, 
+                    directory : str = ''
+                    ) -> None:
+        ''' 
+        Load method for file names in list of names in directory.
+        Return list of algorithms.
+        
+        Use case:
+        algorithms = load_stack(names_list, directory='')
+        
+        names_list is the list of names like ['LGBMClassifier.sav', 'XGBClassifier.sav', 'CatBoostClassifier.sav']
+        directory:
+        '' - read files from current working directory
+        or
+        read files from /some/directory/not/exist/
+        '''
+        return save_load_algorithms.load_stack( names_list, 
+                                                directory=directory, 
+                                                )
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
 
 
 
@@ -743,6 +970,8 @@ class Snaplib:
         stacked_predicts = pd.DataFrame()
         alg_names = []
         increment = 1
+
+
         for algorithm in algorithms_list:
             alg = algorithm[0]
             params = algorithm[1]
@@ -759,7 +988,7 @@ class Snaplib:
                 model.fit(X_train, 
                           y_train, 
                           eval_set=[(X_test, y_test)], 
-                          early_stopping_rounds=200, 
+                          early_stopping_rounds=early_stopping_rounds, 
                           verbose=False)
             else:
                 model = alg(**params)
@@ -1115,10 +1344,6 @@ class Snaplib:
             raise TypeError('early_stopping_rounds must be of int type.')
 
 
-        if y_val is not None:
-            X = pd.concat([X_train, X_val], ignore_index=True)
-            y = pd.concat([y_train, y_val], ignore_index=True)
-
         for alg in algorithms_list:
             alg_name = alg.__class__.__name__[:3]
             if X_val is not None and \
@@ -1132,7 +1357,7 @@ class Snaplib:
                         early_stopping_rounds=early_stopping_rounds, 
                         verbose=verbose)
             else:
-                alg.fit(X, y)
+                alg.fit(X_train, y_train)
 
         return algorithms_list
     
@@ -1746,7 +1971,7 @@ class Snaplib:
         Imputing of missing values (np.nan) in tabular data, not TimeSeries.
 
         Use case:
-        df = Snaplib().recover_data(df, verbose=True, stacking=True)
+        df = Snaplib().recover_data(df, verbose=True, discrete_columns="auto")
         device must be "cpu" or "gpu". Sometime small datasets work faster with cpu.
         if set verbose = if True algorithm runs cross validation tests and print results of tests for decision making.
         discrete_columns = ['col_name_1', 'col_name_2', 'col_name_3', 'etc']
